/****************************************************************************
 *																			*
 * File    : RCFile.c														*
 *																			*
 * Purpose : Sample parser add-in DLL. Supports resource syntax.			*
 *																			*
 * History : Date		 Reason												*
 *			 04-10-15	Created												*
 *																			*
 ****************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <addin.h>
#include <tchar.h>
#include <assert.h>

#define NELEMS(a)  (sizeof(a) / sizeof(a[0]))

// Flags for usCookie.
#define FLAG_COMMENT		0x01
#define FLAG_STRING			0x02
#define FLAG_EXT_COMMENT	0x04
#define FLAG_UNUSED2	0x08	// Free for any use.
#define FLAG_UNUSED3	0x10	// Free for any use.
#define FLAG_UNUSED4	0x20	// Free for any use.
#define FLAG_UNUSED5	0x40	// Free for any use.
#define FLAG_UNUSED6	0x80	// Free for any use.

// Keyword list - must be sorted for IsKeyword().
static PCTSTR apcszKeywords[] = {
	_T("ACCELATORS"),
	_T("ANICURSOR"),
	_T("ANIICON"),
	_T("BEGIN"),
	_T("BITMAP"),
	_T("BLOCK"),
	_T("CAPTION"),
	_T("CLASS"),
	_T("CONTROL"),
	_T("CURSOR"),
	_T("DIALOG"),
	_T("DIALOGEX"),
	_T("DLGINCLUDE"),
	_T("END"),
	_T("FONT"),
	_T("HTML"),
	_T("ICON"),
	_T("LANGUAGE"),
	_T("MANIFEST"),
	_T("MENU"),
	_T("MENUEX"),
	_T("MENUITEM"),
	_T("MESSAGETABLE"),
	_T("PLUGPLAY"),
	_T("POPUP"),
	_T("RCDATA"),
	_T("STRINGTABLE"),
	_T("STYLE"),
	_T("VALUE"),
	_T("VERSIONINFO"),
	_T("VXD"),
};

// Helper macro for assigning color to column position.
/* assert((pos) >= 0 && (pos) < cchText); \ */
#define DEFINE_COLOR(pos,color) \
	if ((pos) >= 0 && (pos) < cchText) \
	if (aPoints != NULL) \
	{ \
		if (*pcPoints == 0 || aPoints[(*pcPoints)-1].iChar <= (pos)) \
		{ \
			aPoints[(*pcPoints)].iChar = (pos);\
			aPoints[(*pcPoints)].iColor = (color);\
			(*pcPoints)++; \
		} \
	}

// Locals.
static HANDLE g_hmod = NULL;

// Function prototypes.
static USHORT Parser(USHORT, PCTSTR, int, ADDIN_PARSE_POINT [], PINT);
static BOOL IsKeyword(PCTSTR [], int, PCTSTR, int);
static BOOL IsNumber(PCTSTR, int);

/****************************************************************************
 *																			*
 * Function: DllMain														*
 *																			*
 * Purpose : DLL entry and exit procedure.									*
 *																			*
 * History : Date		 Reason												*
 *			 04-10-08	Created												*
 *																			*
 ****************************************************************************/

BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
	switch (dwReason)
	{
		case DLL_PROCESS_ATTACH:
			g_hmod = hDLL;
			return TRUE;

		case DLL_PROCESS_DETACH:
			return TRUE;

		default:
			return TRUE;
	}
}

/****************************************************************************
 *																			*
 * Function: AddInMain														*
 *																			*
 * Purpose : Add-in entry and exit procedure.								*
 *																			*
 * History : Date		 Reason												*
 *			 04-10-08	Created												*
 *																			*
 ****************************************************************************/

ADDINAPI BOOL WINAPI AddInMain(HWND hwnd, ADDIN_EVENT eEvent)
{
	switch (eEvent)
	{
		case AIE_APP_CREATE:
		{
			ADDIN_ADD_FILE_TYPE AddFile = {0};

			/* Define a new file type in the IDE */
			AddFile.cbSize = sizeof(AddFile);
			AddFile.pszDescription = "Resource";
			AddFile.pszExtension = "RC";  /* support *.RC files */
			AddFile.pfnParser = Parser;
			if (!AddIn_AddFileType(hwnd, &AddFile))
				return FALSE;

			return TRUE;
		}

		default:
			return TRUE;
	}
}

/****************************************************************************
 *																			*
 * Function: Parser 														*
 *																			*
 * Purpose : Parse source code - for syntax color highlighting. 			*
 *																			*
 * History : Date	   Reason												*
 *			 04-10-08	Created												*
 *																			*
 ****************************************************************************/

static USHORT Parser(USHORT usCookie, PCTSTR pszText, int cchText, ADDIN_PARSE_POINT aPoints[], PINT pcPoints)
{
	BYTE cFoldLevel;
	int iRedefine = 0;
	int iIdentBegin = -1;
	/*BOOL fSeenEnd = FALSE;*/
	/*BOOL fSeenExit = FALSE;*/
	/*BOOL fSeenDeclare = FALSE;*/
	int i;

	// usCookie - (IN) saved state from previous line (zero for first line) - flags (1 byte), folding level (1 byte).
	// pszText  - (IN) pointer to text to parse.
	// cchText  - (IN) length of text to parse (can be zero, for empty lines!).
	// aPoints  - (OUT) array of ADDIN_PARSE_POINT that specifies color changes. Large enough for a color change
	//                  on each character in the largest supported line length (4096 characters).
	// pcPoints - (OUT) number of entries stored in aPoints[].

	// quick exit for empty lines.
	if (cchText == 0)
		return usCookie;

	// break up cookie in low and high byte.
	cFoldLevel = ADDIN_GET_COOKIE_LEVEL(usCookie);
	usCookie = ADDIN_GET_COOKIE_FLAGS(usCookie);

	for (i = 0;; i++)
	{
		if (iRedefine >= 0)
		{
			if (usCookie & (FLAG_COMMENT|FLAG_EXT_COMMENT))
			{
				DEFINE_COLOR(iRedefine, ADDIN_COLOR_COMMENT);
			}
			else if (usCookie & (FLAG_STRING))
			{
				DEFINE_COLOR(iRedefine, ADDIN_COLOR_STRING);
			}
			else
			{
				DEFINE_COLOR(iRedefine, ADDIN_COLOR_TEXT);
			}

			iRedefine = -1;
		}

		// check for end of line.
		if (i == cchText)
			break;

		// inside a single line comment: //
		if (usCookie & FLAG_COMMENT)
		{
			DEFINE_COLOR(i, ADDIN_COLOR_COMMENT);
			usCookie |= FLAG_COMMENT;
			break;
		}

		// inside a string constant: "...."
		if (usCookie & FLAG_STRING)
		{
			// check for end of string constant...
			if (pszText[i] == '"')
			{
				usCookie &= ~FLAG_STRING;
				iRedefine = i + 1;
			}
			continue;
		}

		// inside extended comment: /*....*/
		if (usCookie & FLAG_EXT_COMMENT)
		{
			// check for end of extended comment...
			if (i > 0 && pszText[i] == '/' && pszText[i-1] == '*')
			{
				usCookie &= ~FLAG_EXT_COMMENT;
				iRedefine = i + 1;
			}
			continue;
		}

		// check for single line comment: //
		if (i > 0 && pszText[i] == '/' && pszText[i-1] == '/')
		{
			DEFINE_COLOR(i-1, ADDIN_COLOR_COMMENT);
			usCookie |= FLAG_COMMENT;
			continue;
		}

		// check for a string constant: "...."
		if (pszText[i] == '"')
		{
			DEFINE_COLOR(i, ADDIN_COLOR_STRING);
			usCookie |= FLAG_STRING;
			continue;
		}

		// check for extended comment: /*....*/
		if (i > 0 && pszText[i] == '*' && pszText[i-1] == '/')
		{
			DEFINE_COLOR(i-1, ADDIN_COLOR_COMMENT);
			usCookie |= FLAG_EXT_COMMENT;
			continue;
		}

		// handle identifiers and keywords.
		if (isalnum(pszText[i]) || (iIdentBegin >= 0 && pszText[i] == '_'))
		{
			// remember where the identifier/keyword starts.
			if (iIdentBegin == -1)
				iIdentBegin = i;
		}
		else
		{
			if (iIdentBegin >= 0)
			{
				if (IsKeyword(apcszKeywords, NELEMS(apcszKeywords), pszText + iIdentBegin, i - iIdentBegin))
				{
					DEFINE_COLOR(iIdentBegin, ADDIN_COLOR_KEYWORD);
/*
					// special (hackish) handling of folding points for FUNCTION and SUB.
					if (!fSeenExit && !fSeenDeclare && _tcsnicmp("begin", pszText + iIdentBegin, i - iIdentBegin) == 0)
					{
						if (fSeenEnd && cFoldLevel > 0)  // END FUNCTION - decrease level.
							cFoldLevel--;
						else if (!fSeenEnd && cFoldLevel < 255)  // FUNCTION - increase level.
							cFoldLevel++;
					}
					else
					{
						// remember if we seen END.
						fSeenEnd = _tcsnicmp("end", pszText + iIdentBegin, i - iIdentBegin) == 0;
					}
*/
				}
				else if (IsNumber(pszText + iIdentBegin, i - iIdentBegin))
				{
					DEFINE_COLOR(iIdentBegin, ADDIN_COLOR_NUMBER);
				}

				iRedefine = i;
				iIdentBegin = -1;
			}
		}
	}

	// handle dangling identifiers and keywords.
	if (iIdentBegin >= 0)
	{
		if (IsKeyword(apcszKeywords, NELEMS(apcszKeywords), pszText + iIdentBegin, i - iIdentBegin))
		{
			DEFINE_COLOR(iIdentBegin, ADDIN_COLOR_KEYWORD);
/*
			// special (hackish) handling of folding points for FUNCTION and SUB.
			if (!fSeenExit && !fSeenDeclare && _tcsnicmp("begin", pszText + iIdentBegin, i - iIdentBegin) == 0)
			{
				if (fSeenEnd && cFoldLevel > 0)  // END - decrease level.
					cFoldLevel--;
				else if (!fSeenEnd && cFoldLevel < 255)  // BEGIN - increase level.
					cFoldLevel++;
			}
*/
		}
		else if (IsNumber(pszText + iIdentBegin, i - iIdentBegin))
		{
			DEFINE_COLOR(iIdentBegin, ADDIN_COLOR_NUMBER);
		}
	}

	// return cookie: flags and folding level that the next line will see.
	return ADDIN_MAKE_COOKIE(usCookie & FLAG_EXT_COMMENT,cFoldLevel);
}

/****************************************************************************
 *																			*
 * Function: IsKeyword														*
 *																			*
 * Purpose : Return TRUE if the passed string is a keyword. 				*
 *																			*
 * History : Date		 Reason												*
 *			 04-10-08	Created												*
 *																			*
 ****************************************************************************/

static BOOL IsKeyword(PCTSTR apcszKeywords[], int cKeywords, PCTSTR pszText, int cchText)
{
	int lo = 0;
	int hi = cKeywords-1;

	while (hi >= lo)  /* binary search */
	{
		int this = (lo + hi) / 2;
		int cond = _tcsnicmp(apcszKeywords[this], pszText, cchText);  /* case insensitive */

		if (cond > 0)
			hi = this-1;
		else if (cond < 0)
			lo = this+1;
		else if (apcszKeywords[this][cchText] == '\0')
			return TRUE;
		else if (IsKeyword(apcszKeywords + lo, this - lo, pszText, cchText))
			return TRUE;
		else if (IsKeyword(apcszKeywords + this + 1, hi - this, pszText, cchText))
			return TRUE;
		else
			break;
	}

	return FALSE;
}

/****************************************************************************
 *																			*
 * Function: IsNumber														*
 *																			*
 * Purpose : Return TRUE if the passed string is a number.					*
 *																			*
 * History : Date		 Reason												*
 *			 04-10-08	Created												*
 *																			*
 ****************************************************************************/

static BOOL IsNumber(PCTSTR pszText, int cchText)
{
	// Very *simple* test for numbers - floats, hex constants, are not handled.
	for (int i = 0; i < cchText; i++)
	{
		if (!isdigit(pszText[i]))
			return FALSE;
	}

	return TRUE;
}

